"""Define the scaling report tests."""
import unittest

import numpy as np
import openmdao.api as om

from openmdao.utils.testing_utils import use_tempdirs
from openmdao.utils.assert_utils import assert_near_equal


class ExecCompVOI(om.ExecComp):
    # adds all of its inputs as DVs and all of its outputs as constraints
    def setup(self):
        super().setup()

        # add design vars
        rel2meta = self._var_rel2meta
        for name in sorted(self._var_rel_names['input']):
            meta = rel2meta[name]
            self.add_design_var(name, units=meta['units'])

        # add constraints
        for name in sorted(self._var_rel_names['output']):
            meta = rel2meta[name]
            self.add_constraint(name, lower=0., upper=10., units=meta['units'])


class _Obj(object):
    def __init__(self):
        self.dvs = []
        self.cons = []
        self.objs = []


@use_tempdirs
class TestDriverScalingReport(unittest.TestCase):

    def setup_model(self, mult_exp_range=(-10, 4), nins=1, nouts=1, ncomps=10, shape=1):
        assert nins >= 1
        assert nouts >= 1
        assert ncomps >= 1
        assert mult_exp_range[1] > mult_exp_range[0]

        inidxs = np.arange(nins)
        outidxs = np.arange(nouts)

        expected = _Obj()
        expected.objs.append("objective_comp.out")

        p = om.Problem(allow_post_setup_reorder=False)
        model = p.model
        for icomp in range(ncomps):
            exprs = []
            for iout in range(nouts):

                inperm = np.random.permutation(inidxs)
                imults = "+".join([f"in{i} * 10**{(mult_exp_range[1] - mult_exp_range[0]) * np.random.random() + mult_exp_range[0]}" for
                          i in inperm])
                exprs.append(f"out{iout} = {imults}")

            model.add_subsystem(f"comp{icomp}", ExecCompVOI(exprs, shape=shape, do_coloring=False))

            if icomp == 0:
                # add a comp for the objective
                model.add_subsystem("objective_comp", om.ExecComp("out=inp * 2", shape=shape))
                model.add_objective("objective_comp.out")
                model.connect("comp0.out0", "objective_comp.inp")

            s_ins = sorted(f"in{i}" for i in inidxs)
            expected.dvs.extend(f"comp{icomp}.{n}" for n in s_ins)
            s_outs = sorted(f"out{i}" for i in outidxs)
            expected.cons.extend(f"comp{icomp}.{n}" for n in s_outs)

        p.setup()
        return p, expected

    def _check_data(self, data, expected):
        objs = data['oflabels'][0]
        cons = data['oflabels'][1:]
        self.assertEqual(expected.objs, [objs])
        self.assertEqual(expected.cons, cons)
        self.assertEqual(expected.dvs, data['wrtlabels'])
        self.assertEqual(0, len(data['linear']['oflabels']))
        self.assertEqual(len(expected.objs), len(data['obj_table']))
        self.assertEqual(len(expected.cons), len(data['con_table']))
        self.assertEqual(len(expected.dvs), len(data['dv_table']))
        for dvrow in data['dv_table']:
            if dvrow['size'] > 1:
                self.assertEqual(dvrow['size'], len(dvrow['_children']))
                for chrow in dvrow['_children']:
                    self.assertEqual('', chrow['size'])
                    self.assertTrue('_children' not in chrow)

        for conrow in data['con_table']:
            if conrow['size'] > 1:
                self.assertEqual(conrow['size'], len(conrow['_children']))
                for chrow in conrow['_children']:
                    self.assertEqual('', chrow['size'])
                    self.assertTrue('_children' not in chrow)

        for objrow in data['obj_table']:
            if objrow['size'] > 1:
                self.assertEqual(objrow['size'], len(objrow['_children']))
                for chrow in objrow['_children']:
                    self.assertEqual('', chrow['size'])
                    self.assertTrue('_children' not in chrow)

    def test_40x40_in1_out1(self):
        p, expected = self.setup_model(ncomps=40)
        p.final_setup()
        # compute dict totals to make sure we handle that properly
        p.driver._compute_totals()
        data = p.driver.scaling_report(show_browser=False)
        self._check_data(data, expected)

    def test_40x40_in4_out4(self):
        p, expected = self.setup_model(nins=4, nouts=4, ncomps=10)
        p.final_setup()
        data = p.driver.scaling_report(show_browser=False)
        self._check_data(data, expected)

    def test_40x40_in4_out4_shape10(self):
        p, expected = self.setup_model(nins=4, nouts=4, ncomps=10, shape=10)
        p.final_setup()
        data = p.driver.scaling_report(show_browser=False)
        self._check_data(data, expected)

    def test_in100out4_shape10(self):
        p, expected = self.setup_model(nins=100, nouts=4, ncomps=1, shape=10)
        p.final_setup()
        data = p.driver.scaling_report(show_browser=False)
        self._check_data(data, expected)

    def test_in4out100_shape10(self):
        p, expected = self.setup_model(nins=4, nouts=100, ncomps=1, shape=10)
        p.final_setup()
        data = p.driver.scaling_report(show_browser=False)
        self._check_data(data, expected)

    def test_big_subjac(self):
        # this gets slow with larger sizes, e.g. shape=1000 takes > 20 sec to
        # show/hide table rows or show a subjac in the browser
        p, expected = self.setup_model(nins=4, nouts=4, ncomps=1, shape=100)
        p.final_setup()
        data = p.driver.scaling_report(show_browser=False)
        self._check_data(data, expected)


@use_tempdirs
class TestDriverScalingReport2(unittest.TestCase):

    def test_unconstrained(self):
        from openmdao.test_suite.components.paraboloid import Paraboloid

        # build the model
        prob = om.Problem()
        indeps = prob.model.add_subsystem('indeps', om.IndepVarComp())
        indeps.add_output('x', 3.0)
        indeps.add_output('y', -4.0)

        prob.model.add_subsystem('paraboloid', Paraboloid())

        prob.model.connect('indeps.x', 'paraboloid.x')
        prob.model.connect('indeps.y', 'paraboloid.y')

        # setup the optimization
        prob.driver = om.ScipyOptimizeDriver(optimizer='COBYLA')

        prob.model.add_design_var('indeps.x', lower=-50, upper=50)
        prob.model.add_design_var('indeps.y', lower=-50, upper=50)
        prob.model.add_objective('paraboloid.f_xy', index=0)

        prob.setup()
        prob.run_driver()

        # minimum value
        assert_near_equal(prob['paraboloid.f_xy'], -27.33333, 1e-6)

        # location of the minimum
        assert_near_equal(prob['indeps.x'], 6.6667, 1e-4)
        assert_near_equal(prob['indeps.y'], -7.33333, 1e-4)

        # just make sure this doesn't raise an exception
        prob.driver.scaling_report(show_browser=False)


@use_tempdirs
class TestDriverScalingReport3(unittest.TestCase):
    def test_setup_message(self):
        x_train = np.arange(0., 10.)
        y_train = np.arange(10., 20.)
        z_train = x_train ** 2 + y_train ** 2

        p = om.Problem()
        p.model = model = om.Group()

        params = om.IndepVarComp()
        params.add_output('x', val=0.)
        params.add_output('y', val=0.)

        model.add_subsystem('params', params, promotes=['*'])

        sm = om.MetaModelUnStructuredComp(default_surrogate=om.ResponseSurface())
        sm.add_input('x', val=0.)
        sm.add_input('y', val=0.)
        sm.add_output('z', val=0.)

        sm.options['train_x'] = x_train
        sm.options['train_y'] = y_train
        sm.options['train_z'] = z_train

        # With or without the line below does not matter
        # Only when method is set to fd, then RuntimeWarning disappears
        sm.declare_partials('*', '*', method='exact')

        model.add_subsystem('sm', sm, promotes=['*'])

        model.add_design_var('x', lower=0., upper=10.)
        model.add_design_var('y', lower=0., upper=10.)
        model.add_objective('z')

        p.setup()

        with self.assertRaises(RuntimeError) as cm:
            p.driver.scaling_report()

        msg = "Either 'run_model' or 'final_setup' must be called before the scaling report can be generated."
        self.assertEqual(str(cm.exception), msg)

        # Now, make sure it runs run_model automatically as long as we final_setup.
        p.final_setup()
        p.driver.scaling_report(show_browser=False)


@use_tempdirs
class TestDiscreteScalingReport(unittest.TestCase):

    def test_scaling_report(self):
        class ParaboloidDiscrete(om.ExplicitComponent):

            def setup(self):
                self.add_discrete_input('x', val=10)
                self.add_discrete_input('y', val=0)
                self.add_discrete_output('f_xy', val=0)

            def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
                x = discrete_inputs['x']
                y = discrete_inputs['y']
                f_xy = (x - 3.0)**2 + x * y + (y + 4.0)**2 - 3.0
                discrete_outputs['f_xy'] = int(f_xy)

        prob = om.Problem()
        model = prob.model

        # Add independent variables
        indeps = model.add_subsystem('indeps', om.IndepVarComp(), promotes=['*'])
        indeps.add_discrete_output('x', 4)
        indeps.add_discrete_output('y', 3)

        # Add components
        model.add_subsystem('parab', ParaboloidDiscrete(), promotes=['*'])

        # Specify design variable range and objective
        model.add_design_var('x')
        model.add_design_var('y')
        model.add_objective('f_xy')

        samples = [
            [('x', 5), ('y', 1)],
            [('x', 3), ('y', 6)],
            [('x', -1), ('y', 3)],
        ]

        # Setup driver
        prob.driver = om.DOEDriver(om.ListGenerator(samples))

        # run driver
        prob.setup()
        prob.run_driver()
        prob.cleanup()

        # generate scaling report
        prob.driver.scaling_report(show_browser=False)

    def test_bug_2494(self):
        # see Issue #2494, bug in scaling_report was causing damaging side effect
        class OMGroup(om.Group):
            def setup(self):
                ivc = self.add_subsystem("indep_vars", om.IndepVarComp(), promotes=["*"])
                ivc.add_output("length", val=0.0, desc="Length")
                ivc.add_output("width",val=0.0, desc="Width")

                self.add_subsystem("comp",OMComponent(),  promotes=["*"])

        class OMComponent(om.ExplicitComponent):
            def setup(self):
                self.add_discrete_input("definition", val=1, desc="Flag")
                self.add_input("length", val=0.0, desc="Length")
                self.add_input("width", val=0.0, desc="Width")

                self.add_output("area", val=0.0, desc="Area")

            def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
                if discrete_inputs['definition'] == 0:
                    outputs["area"] = inputs['length'] * inputs['width']
                else:
                    outputs["area"] = 2.* inputs['length'] * inputs['width']

        myopt = om.Problem(model=OMGroup(), driver=om.ScipyOptimizeDriver())

        myopt.model.approx_totals(method="fd")
        myopt.model.add_objective("area")
        myopt.model.add_design_var("length",lower=0,upper=10)

        myopt.setup()

        myopt["width"] = 3.
        myopt["length"] = 2.
        myopt.run_driver()

        # generate scaling report
        myopt.driver.scaling_report(show_browser=False)

        # verify access to discrete value
        self.assertEqual(myopt["definition"], 1)


if __name__ == '__main__':
    unittest.main()



"""Tests the `debug_print` option for Nonlinear solvers."""

import sys

import unittest
from packaging.version import Version
from io import StringIO

import numpy as np

import openmdao.api as om
from openmdao.test_suite.scripts.circuit_analysis import Circuit

from openmdao.utils.general_utils import run_model, printoptions
from openmdao.utils.testing_utils import use_tempdirs

try:
    from parameterized import parameterized
except ImportError:
    from openmdao.utils.assert_utils import SkipParameterized as parameterized

nonlinear_solvers = [
    om.NonlinearBlockGS,
    om.NonlinearBlockJac,
    om.NewtonSolver,
    om.BroydenSolver
]

@use_tempdirs
class TestNonlinearSolvers(unittest.TestCase):
    def setUp(self):
        # iteration coordinate, file name and variable data are common for all tests
        coord = 'rank0:root._solve_nonlinear|0|NLRunOnce|0|circuit._solve_nonlinear|0'
        self.filename = 'solver_errors.0.out'

        self.expected_data = '\n'.join([
            "",
            "# Inputs and outputs at start of iteration '%s':" % coord,
            "",
            "# nonlinear inputs",
            "{'circuit.D1.V_in': array([ 1.+0.j]),",
            " 'circuit.D1.V_out': array([ 0.+0.j]),",
            " 'circuit.R1.V_in': array([ 1.+0.j]),",
            " 'circuit.R1.V_out': array([ 0.+0.j]),",
            " 'circuit.R2.V_in': array([ 1.+0.j]),",
            " 'circuit.R2.V_out': array([ 1.+0.j]),",
            " 'circuit.n1.I_in:0': array([ 0.1+0.j]),",
            " 'circuit.n1.I_out:0': array([ 1.+0.j]),",
            " 'circuit.n1.I_out:1': array([ 1.+0.j]),",
            " 'circuit.n2.I_in:0': array([ 1.+0.j]),",
            " 'circuit.n2.I_out:0': array([ 1.+0.j])}",
            "",
            "# nonlinear outputs",
            "{'circuit.D1.I': array([ 1.+0.j]),",
            " 'circuit.R1.I': array([ 1.+0.j]),",
            " 'circuit.R2.I': array([ 1.+0.j]),",
            " 'circuit.n1.V': array([ 10.+0.j]),",
            " 'circuit.n2.V': array([ 0.001+0.j])}",
            ""
        ])

    @parameterized.expand([
        [solver.__name__, solver] for solver in nonlinear_solvers
    ])
    def test_solver_debug_print(self, name, solver):
        p = om.Problem()
        model = p.model

        model.add_subsystem('ground', om.IndepVarComp('V', 0., units='V'))
        model.add_subsystem('source', om.IndepVarComp('I', 0.1, units='A'))
        model.add_subsystem('circuit', Circuit())

        model.connect('source.I', 'circuit.I_in')
        model.connect('ground.V', 'circuit.Vg')

        p.setup()
        nl = model.circuit.nonlinear_solver = solver()

        nl.options['debug_print'] = True
        nl.options['err_on_non_converge'] = True

        if name == 'NonlinearBlockGS':
            nl.options['use_apply_nonlinear'] = True

        if name == 'NewtonSolver':
            nl.options['solve_subsystems'] = True

        # suppress solver output for test
        nl.options['iprint'] = model.circuit.linear_solver.options['iprint'] = -1

        # For Broydensolver, don't calc Jacobian
        try:
            nl.options['compute_jacobian'] = False
        except KeyError:
            pass

        # set some poor initial guesses so that we don't converge
        p['circuit.n1.V'] = 10.
        p['circuit.n2.V'] = 1e-3

        opts = {}
        # formatting has changed in numpy 1.14 and beyond.
        if Version(np.__version__) >= Version("1.14"):
            opts["legacy"] = '1.13'

        with printoptions(**opts):
            # run the model and check for expected output file
            output = run_model(p, ignore_exception=True)

        expected_output = '\n'.join([
            self.expected_data,
            "Inputs and outputs at start of iteration "
            "have been saved to '%s'.\n" % self.filename
        ])

        self.assertEqual(output, expected_output)

        with open(self.filename, 'r') as f:
            self.assertEqual(f.read(), self.expected_data)

        # setup & run again to make sure there is no error due to existing file
        p.setup()
        with printoptions(**opts):
            run_model(p, ignore_exception=False)

    def test_solver_debug_print_feature(self):

        p = om.Problem()
        model = p.model

        model.add_subsystem('circuit', Circuit())

        p.setup()

        nl = model.circuit.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)

        nl.options['iprint'] = 2
        nl.options['debug_print'] = True
        nl.options['err_on_non_converge'] = True

        # set some poor initial guesses so that we don't converge
        p.set_val('circuit.I_in', 0.1, units='A')
        p.set_val('circuit.Vg', 0.0, units='V')
        p.set_val('circuit.n1.V', 10.)
        p.set_val('circuit.n2.V', 1e-3)

        opts = {}
        # formatting has changed in numpy 1.14 and beyond.
        if Version(np.__version__) >= Version("1.14"):
            opts["legacy"] = '1.13'

        with printoptions(**opts):
            # run the model
            try:
                p.run_model()
            except om.AnalysisError:
                pass

        with open(self.filename, 'r') as f:
            self.assertEqual(f.read(), self.expected_data)

    def test_solver_debug_print_multi_run_scenario(self):

        p = om.Problem()
        model = p.model

        model.add_subsystem('circuit', Circuit())

        p.setup()

        nl = model.circuit.nonlinear_solver = om.NewtonSolver(solve_subsystems=False)

        nl.options['iprint'] = 2
        nl.options['debug_print'] = True
        nl.options['err_on_non_converge'] = False

        # set some poor initial guesses so that we don't converge
        p.set_val('circuit.I_in', 0.1, units='A')
        p.set_val('circuit.Vg', 0.0, units='V')
        p.set_val('circuit.n1.V', 10.)
        p.set_val('circuit.n2.V', 1e-3)

        p.run_model()
        p.set_solver_print(level=-1)
        output = run_model(p)

        # Should be empty since solver debugging printing was turned off
        self.assertEqual(output, '')

@use_tempdirs
class TestNonlinearSolversIsolated(unittest.TestCase):
    """
    This test needs to run isolated to preclude interactions in the underlying
    `warnings` module that is used to raise the singular entry error.
    """
    ISOLATED = True

    def test_debug_after_raised_error(self):
        prob = om.Problem()
        model = prob.model

        comp = om.IndepVarComp()
        comp.add_output('dXdt:TAS', val=1.0)
        comp.add_output('accel_target', val=2.0)
        model.add_subsystem('des_vars', comp, promotes=['*'])

        teg = model.add_subsystem('thrust_equilibrium_group', subsys=om.Group())
        teg.add_subsystem('dynamics', om.ExecComp('z = 2.0*thrust'), promotes=['*'])

        thrust_bal = om.BalanceComp()
        thrust_bal.add_balance(name='thrust', val=1207.1, lhs_name='dXdt:TAS',
                               rhs_name='accel_target', eq_units='m/s**2', lower=-10.0, upper=10000.0)

        teg.add_subsystem(name='thrust_bal', subsys=thrust_bal,
                          promotes_inputs=['dXdt:TAS', 'accel_target'],
                          promotes_outputs=['thrust'])

        teg.linear_solver = om.DirectSolver()

        teg.nonlinear_solver = om.NewtonSolver()
        teg.nonlinear_solver.options['solve_subsystems'] = True
        teg.nonlinear_solver.options['max_sub_solves'] = 1
        teg.nonlinear_solver.options['atol'] = 1e-4
        teg.nonlinear_solver.options['debug_print'] = True

        prob.setup()
        prob.set_solver_print(level=0)

        stdout = sys.stdout
        strout = StringIO()
        sys.stdout = strout

        with self.assertRaises(RuntimeError) as cm:
            prob.run_model()

        sys.stdout = stdout

        output = strout.getvalue()
        target = "'thrust_equilibrium_group.thrust_bal.thrust'"
        self.assertTrue(target in output, msg=target + "NOT FOUND IN" + output)

        # Make sure exception is unchanged.
        expected_msg = "Singular entry found in 'thrust_equilibrium_group' <class Group> for row associated with state/residual 'thrust' ('thrust_equilibrium_group.thrust_bal.thrust') index 0."
        self.assertEqual(expected_msg, str(cm.exception))


if __name__ == "__main__":
    unittest.main()



"""Unit Tests for n2_viewer"""
import unittest

import numpy as np

import openmdao.api as om
from openmdao.test_suite.components.paraboloid import Paraboloid
from openmdao.visualization.n2_viewer.n2_viewer import _get_viewer_data
from openmdao.utils.testing_utils import use_tempdirs
from openmdao.test_suite.components.sellar import SellarDerivativesGrouped, SellarProblem


@use_tempdirs
class TestN2Viewer(unittest.TestCase):

    def setUp(self):

        self.filename = "cases.sql"
        self.recorder = om.SqliteRecorder(self.filename)

        # design variable values as generated by Placket Burman DOE generator
        self.doe_list = [
            [('x', 0.), ('y', 0.)],
            [('x', 1.), ('y', 0.)],
            [('x', 0.), ('y', 1.)],
            [('x', 1.), ('y', 1.)]
        ]

        self.driver_case = "rank0:DOEDriver_List|3"
        self.problem_case = "rank0:DOEDriver_List|3|root._solve_nonlinear|3"

    def test_driver_case(self):
        prob = om.Problem()
        model = prob.model

        model.add_subsystem('p1', om.IndepVarComp('x', 0.0), promotes=['x'])
        model.add_subsystem('p2', om.IndepVarComp('y', 0.0), promotes=['y'])
        model.add_subsystem('comp', Paraboloid(), promotes=['x', 'y', 'f_xy'])

        model.add_design_var('x', lower=0.0, upper=1.0)
        model.add_design_var('y', lower=0.0, upper=1.0)
        model.add_objective('f_xy')

        prob.driver = om.DOEDriver(self.doe_list)
        prob.driver.add_recorder(self.recorder)
        prob.driver.recording_options['includes'] = ['*']

        prob.setup()
        prob.run_driver()
        prob.cleanup()

        data_dict = _get_viewer_data(prob.get_outputs_dir() / self.filename, case_id=self.driver_case)

        vals = data_dict['tree']['children'][2]['children']
        x_val = vals[0]['val']
        y_val = vals[1]['val']
        f_xy_val = vals[2]['val']

        self.assertEqual(x_val, np.array([1.]))
        self.assertEqual(y_val, np.array([1.]))
        self.assertEqual(f_xy_val, np.array([27.]))

    def test_driver_case_na(self):
        prob = om.Problem()
        model = prob.model

        model.add_subsystem('p1', om.IndepVarComp('x', 0.0), promotes=['x'])
        model.add_subsystem('p2', om.IndepVarComp('y', 0.0), promotes=['y'])
        model.add_subsystem('comp', Paraboloid(), promotes=['x', 'y', 'f_xy'])

        model.add_design_var('x', lower=0.0, upper=1.0)
        model.add_design_var('y', lower=0.0, upper=1.0)
        model.add_objective('f_xy')

        prob.driver = om.DOEDriver(self.doe_list)
        prob.driver.add_recorder(self.recorder)

        prob.setup()
        prob.run_driver()
        prob.cleanup()

        data_dict = _get_viewer_data(prob.get_outputs_dir() / self.filename, case_id=self.driver_case)

        vals = data_dict['tree']['children'][2]['children']
        x_val = vals[0]['val']
        y_val = vals[1]['val']
        f_xy_val = vals[2]['val']

        self.assertEqual(x_val, "N/A")
        self.assertEqual(y_val, "N/A")
        self.assertEqual(f_xy_val, np.array([27.]))

    def test_index_number(self):
        prob = om.Problem()
        model = prob.model

        model.add_subsystem('p1', om.IndepVarComp('x', 0.0), promotes=['x'])
        model.add_subsystem('p2', om.IndepVarComp('y', 0.0), promotes=['y'])
        model.add_subsystem('comp', Paraboloid(), promotes=['x', 'y', 'f_xy'])

        model.add_design_var('x', lower=0.0, upper=1.0)
        model.add_design_var('y', lower=0.0, upper=1.0)
        model.add_objective('f_xy')

        prob.driver = om.DOEDriver(self.doe_list)
        prob.driver.add_recorder(self.recorder)

        prob.setup()
        prob.run_driver()
        prob.cleanup()

        data_dict = _get_viewer_data(prob.get_outputs_dir() / self.filename, case_id=3)

        vals = data_dict['tree']['children'][2]['children']
        x_val = vals[0]['val']
        y_val = vals[1]['val']
        f_xy_val = vals[2]['val']

        self.assertEqual(x_val, "N/A")
        self.assertEqual(y_val, "N/A")
        self.assertEqual(f_xy_val, np.array([27.]))

    def test_root_case(self):
        prob = om.Problem()
        model = prob.model

        model.add_subsystem('p1', om.IndepVarComp('x', 0.0), promotes=['x'])
        model.add_subsystem('p2', om.IndepVarComp('y', 0.0), promotes=['y'])
        model.add_subsystem('comp', Paraboloid(), promotes=['x', 'y', 'f_xy'])

        model.add_design_var('x', lower=0.0, upper=1.0)
        model.add_design_var('y', lower=0.0, upper=1.0)
        model.add_objective('f_xy')

        recorder = om.SqliteRecorder("cases.sql")
        prob.driver = om.DOEDriver(self.doe_list)
        prob.driver.add_recorder(recorder)
        prob.model.add_recorder(recorder)

        prob.setup()
        prob.run_driver()
        prob.cleanup()

        data_dict = _get_viewer_data(prob.get_outputs_dir() / "cases.sql", case_id=self.problem_case)

        vals = data_dict['tree']['children'][2]['children']
        x_val = vals[0]['val']
        y_val = vals[1]['val']
        f_xy_val = vals[2]['val']

        self.assertEqual(x_val, np.array([1.]))
        self.assertEqual(y_val, np.array([1.]))
        self.assertEqual(f_xy_val, np.array([27.]))

    def test_driver_root_index_case(self):
        prob = om.Problem()
        model = prob.model

        model.add_subsystem('p1', om.IndepVarComp('x', 0.0), promotes=['x'])
        model.add_subsystem('p2', om.IndepVarComp('y', 0.0), promotes=['y'])
        model.add_subsystem('comp', Paraboloid(), promotes=['x', 'y', 'f_xy'])

        model.add_design_var('x', lower=0.0, upper=1.0)
        model.add_design_var('y', lower=0.0, upper=1.0)
        model.add_objective('f_xy')

        recorder = om.SqliteRecorder("cases.sql")
        prob.driver = om.DOEDriver(self.doe_list)
        prob.driver.add_recorder(recorder)
        prob.model.add_recorder(recorder)

        prob.setup()
        prob.run_driver()
        prob.cleanup()

        data_dict = _get_viewer_data(prob.get_outputs_dir() / "cases.sql", case_id=6)

        vals = data_dict['tree']['children'][2]['children']
        x_val = vals[0]['val']
        y_val = vals[1]['val']
        f_xy_val = vals[2]['val']

        self.assertEqual(x_val, np.array([1.]))
        self.assertEqual(y_val, np.array([1.]))
        self.assertEqual(f_xy_val, np.array([27.]))

    def test_root_index_case(self):
        prob = om.Problem()
        model = prob.model

        model.add_subsystem('p1', om.IndepVarComp('x', 0.0), promotes=['x'])
        model.add_subsystem('p2', om.IndepVarComp('y', 0.0), promotes=['y'])
        model.add_subsystem('comp', Paraboloid(), promotes=['x', 'y', 'f_xy'])

        model.add_design_var('x', lower=0.0, upper=1.0)
        model.add_design_var('y', lower=0.0, upper=1.0)
        model.add_objective('f_xy')

        recorder = om.SqliteRecorder("cases.sql")
        prob.driver = om.DOEDriver(self.doe_list)
        prob.model.add_recorder(recorder)

        prob.setup()
        prob.run_driver()
        prob.cleanup()

        data_dict = _get_viewer_data(prob.get_outputs_dir() / "cases.sql", case_id=3)

        vals = data_dict['tree']['children'][2]['children']
        x_val = vals[0]['val']
        y_val = vals[1]['val']
        f_xy_val = vals[2]['val']

        self.assertEqual(x_val, np.array([1.]))
        self.assertEqual(y_val, np.array([1.]))
        self.assertEqual(f_xy_val, np.array([27.]))

    def test_auto_ivc_case(self):
        prob = SellarProblem(SellarDerivativesGrouped)

        prob.driver = om.ScipyOptimizeDriver(tol=1e-9, disp=False)
        prob.driver.add_recorder(self.recorder)

        prob.setup()
        prob.run_driver()
        prob.cleanup()

        cr = om.CaseReader(prob.get_outputs_dir() / self.filename)
        first_case = cr.list_cases(out_stream=None)[0]

        data_dict = _get_viewer_data(prob.get_outputs_dir() / self.filename, case_id=first_case)

        vals = data_dict['tree']['children'][0]['children']
        ivc_0_val = vals[0]['val']
        ivc_1_val = vals[1]['val']

        self.assertEqual(ivc_0_val[0], 5.)
        self.assertEqual(ivc_0_val[1], 2.)
        self.assertEqual(ivc_1_val, 1.)

    def test_problem_case(self):
        prob = om.Problem()
        model = prob.model

        model.add_subsystem('p1', om.IndepVarComp('x', 0.0), promotes=['x'])
        model.add_subsystem('p2', om.IndepVarComp('y', 0.0), promotes=['y'])
        model.add_subsystem('comp', Paraboloid(), promotes=['x', 'y', 'f_xy'])

        model.add_design_var('x', lower=0.0, upper=1.0)
        model.add_design_var('y', lower=0.0, upper=1.0)
        model.add_objective('f_xy')

        prob.driver = om.DOEDriver(self.doe_list)
        prob.add_recorder(self.recorder)
        prob.recording_options['record_inputs'] = True

        prob.setup()
        prob.run_driver()
        prob.record('final')
        prob.cleanup()

        data_dict = _get_viewer_data(prob.get_outputs_dir() / self.filename, case_id='final')

        vals = data_dict['tree']['children'][2]['children']
        x_val = vals[0]['val']
        y_val = vals[1]['val']
        f_xy_val = vals[2]['val']

        self.assertEqual(x_val, np.array([1.]))
        self.assertEqual(y_val, np.array([1.]))
        self.assertEqual(f_xy_val, np.array([27.]))


if __name__ == "__main__":
    unittest.main()




"""
CaseReader factory function.
"""
from openmdao.recorders.sqlite_reader import SqliteCaseReader


def CaseReader(filename, pre_load=True, metadata_filename=None):
    """
    Return a CaseReader for the given file.

    Parameters
    ----------
    filename : str or Path
        A path to the recorded file.
        Currently only sqlite database files recorded via SqliteRecorder are supported.
    pre_load : bool
        If True, load all the data into memory during initialization.
    metadata_filename : str
        For separate metadata from parallel runs, the metadata database filename.

    Returns
    -------
    BaseCaseReader
        An instance of a CaseReader.
    """
    return SqliteCaseReader(filename, pre_load, metadata_filename)



"""
Convert a Matlab matrix file (like those found at sparse.tamu.edu) to scipy sparse .npz file.
"""

import numpy as np
import scipy.io
from scipy.sparse import save_npz


def matlab2npy(mlfile):
    d = scipy.io.loadmat(mlfile)

    for i in [1, 2, 0]:
        a = d["Problem"][0][0][i]
        try:
            a.data = np.abs(a.data)
        except Exception:
            continue
        else:
            break
    else:
        raise RuntimeError("couldn't find the matrix!!!")


    nrows, ncols = a.shape

    print("Nonzeros:", np.count_nonzero(a.data), "of", nrows * ncols)
    print("Max:", a.data.max())
    print("Min nonzero:", a.data[a.data > 0.].min())
    print("Shape:", a.shape)

    return a


if __name__ == '__main__':
    import sys
    import os

    a = matlab2npy(sys.argv[1])
    base = os.path.splitext(sys.argv[1])[0]
    save_npz('{}.npz'.format(base), a)
